#!/bin/bash
#* Licensed Materials - Property of IBM
#*
#* OCO Source Materials
#*
#* (C) Copyright IBM Corp. 2011 All Rights Reserved
#*
# hostvolattach.sh [-l] [-m] [--ids <WWID>:<LUN>~<address> ...]
#   version: 1.1b [ 1/9/2012]
#   author:  csky
#
# !Script only supported on RHEL 6!
#
#    Description: Scan scsi bus HBA's for fresh set of volumes. Optionally look
#                 for a specific set. Dump out info for the set or all disks found.
#                 Call a fixing script if any of the specific set of volumes are
#                 not located
#    Options:
#        -l    long (verbose)
#        -m    multi-path enabled
#       --ids   All arguments following this option are treated as a specific volume
#               reference, consisting of a WWID (DeviceID) and LUN combination
#               separated by a ':'. If specified, then only those devices are fully
#               processed. Missing stable paths can be fixed if the -f option is
#               specified in this case. If the LUN is unknown, pass -1. <address>
#               is the exteral named path to the storage (not the host view).
#------------------------------------
# 

vmcdir="/var/opt/ibm/vmc"
fixlogfile="${vmcdir}/outsancheck.trace"
alogfile="${vmcdir}/outattach.trace"
argids=N
long=false
if [ "$1" = "-l" -o "$2" = "-l" ]
then
	long=true
fi
multipath=N
if [ "$1" = "-m" -o "$2" = "-m" ]
then
	multipath=Y
fi

if [ "$1" = "--ids" ]
then
	shift 1
	argids=Y
else
	if [ "$2" = "--ids" ]
	then
		shift 2
		argids=Y
	else
		if [ "$3" = "--ids" ]
		then
			shift 3
			argids=Y
		fi
	fi
fi

# This folder should exist since normally the scripts are contained in it.
if [ -d $vmcdir ]
then
	if [ "$long" = "true" ]
	then
		echo "VMC dir exists."
	fi
else
	mkdir -p $vmcdir
	rc=$?
	if [ "$long" = "true" ]
	then
		echo "Created directory $vmcdir. rc=$rc"
	fi
fi

if [ -f $alogfile ]
then
	ccrc=$(find $alogfile -size +1000)
	if [ "$ccrc" = "$alogfile" ]
	then
		echo "Cleanup old log file (wrap to backup)."
		mv -f $alogfile "${alogfile}.1"
	fi
fi

date >> $alogfile
echo "long=$long   argids=$argids   multipath=$multipath" >> $alogfile
touch $vmcdir/pvmap

if [ "$argids" = "Y" ]
then
	volumes="$@"
	echo "Volumes= $volumes" >> $alogfile
	if [ "$long" = "true" ]
	then
		echo "Volumes= $volumes"
	fi
fi

# Check for FC transport
hosts=""
if [ -d /sys/class/fc_transport ]
then
	echo "Detection of FC transport folder." >> $alogfile
	if [ "$long" = "true" ]
	then
		echo "Detection of FC transport folder."
	fi
	
	# loop over transport combinations (HBAs, channels, controllers)
	for fcdir in $(ls -1 /sys/class/fc_transport)
	do
		target=$(echo $fcdir | cut -d : -f 1)
		channel=$(echo $fcdir | cut -d : -f 2)
		controller=$(echo $fcdir | cut -d : -f 3)
		hostnum=${target#target}
		host=host${target#target}
		hosts=${hosts%%$host}
		if [ -z "$hosts" ]
		then
			hosts=$host
		else
			hosts="$hosts,$host"
		fi
		portid=$(cat /sys/class/fc_transport/$fcdir/port_id)
		portwwn=$(cat /sys/class/fc_transport/$fcdir/node_name)
		hportid=$(cat /sys/class/fc_host/$host/port_id)
		hportwwn=$(cat /sys/class/fc_host/$host/port_name)
		echo "  FC transport found: $fcdir" >> $alogfile
		if [ "$long" = "true" ]
		then
			echo "FC transport found: $fcdir"
		fi
	# done with FC transports
	done
fi

# Compute list of HBAs
if [ -z "$hosts" ]
then
	echo "FC transport(s) not detected. Scan all FC HBAs to kickstart process."
	for fchba in $(ls -1 /sys/class/fc_host)
	do
		echo "FC HBA found: $fchba" >> $alogfile
		if [ "$long" = "true" ]
		then
			echo "FC HBA found: $fchba"
		fi
		if [ -z "$hosts" ]
		then
			hosts=$fchba
		else
			hosts="$hosts,$fchba"
		fi
	done
fi

if [ -z "$hosts" ]
then
	echo "This endpoint does not appear to be attached to a Fibre Channel network. Exiting."
	echo "This endpoint does not appear to be attached to a Fibre Channel network. Exiting." >> $alogfile
	date >> $alogfile
	exit 1
fi

# Scan SCSI bus for FC hosts (HBAs)
for rescan in $(echo $hosts | sed s/,/\\n/g)
do
	echo "   echo '- - -'  > /sys/class/scsi_host/$rescan/scan" >> $alogfile
	if [ "$long" = "true" ]
	then
		echo "   echo '- - -'  > /sys/class/scsi_host/$rescan/scan"
	fi
	echo "- - -" > /sys/class/scsi_host/$rescan/scan
done

if [ -f $fixlogfile ]
then
	ccrc=$(find $fixlogfile -size +2000)
	if [ "$ccrc" = "$fixlogfile" ]
	then
		if [ "$long" = "true" ]
		then
			echo "Cleanup old log file (wrap to backup)."
		fi
		echo "Cleanup old log file (wrap to backup)." >> $alogfile
		mv -f $fixlogfile "${fixlogfile}.1"
	fi
fi

# trying id-enabled mode here with recent script changes
fixmode="id-enabled"
firsttime="true"
let innertries=1
# Outer loop - fix mode
while [ "$fixmode" != "done" ]
do
	echo "Begin outer loop. Fixmode=$fixmode" >> $alogfile
	if [ "$long" = "true" ]
	then
		echo "Begin outer loop. Fixmode=$fixmode"
	fi
	# Wait a short time for kernel and udev eventing to populate block device links
	echo "Pause 5..." >> $alogfile
	if [ "$long" = "true" ]
	then
		echo "Pause 5..." 
	fi
	sleep 5
	if [ "$argids" = "N" ]
	then
		# Specific volumes not passed in, so process all current volumes
		idlistfile=$(mktemp $vmcdir/idlist.XXXXXXXX)
		#for sysblk in $(ls -1 -d /sys/block/sd*)
		for fcdir in $(ls -1 /sys/class/fc_transport)
		do
			target=$(echo $fcdir | cut -d : -f 1)
			channel=$(echo $fcdir | cut -d : -f 2)
			controller=$(echo $fcdir | cut -d : -f 3)
			hostnum=${target#target}
			#echo "Dir: /sys/class/fc_transport/$fcdir/device"
			for file in $(ls -1 /sys/class/fc_transport/$fcdir/device | egrep "^$hostnum:$channel:$controller:")
			do
				dlink=/sys/class/fc_transport/$fcdir/device/$file
				if [ -d $dlink/block ]
				then
					blockdev=$(ls $dlink/block)
					#echo "$file: $blockdev"
					devblockpath=/dev/$blockdev
					scsiid=$(scsi_id --page=0x83 --whitelisted --device=$devblockpath | cut -d ' ' -f 1)
					scsiid=${scsiid##[ ]*}
					if [ ${#scsiid} -gt 2 ]
					then
						volid=${scsiid:1}
						volid=${volid##[ ]*}
						if [ -n "$volid" ]
						then
							# Avoid duplicates so just take the first device with the wwid
							ogrep=$(grep $volid $idlistfile)
							rc=$?
							if [ "$rc" != "0" ]
							then
								pvdevname=$(awk -v ss="$volid" '$0 ~ ss {print $13}' $vmcdir/pvmap | head -n 1)
								pvlun=$(awk -v ss="$volid" '$0 ~ ss {print $5}' $vmcdir/pvmap | head -n 1)
								if [ -n "$pvdevname" ]
								then
									devblockpath=$pvdevname
								fi
								if [ -z "$pvlun" ]
								then
									pvlun=-1
								fi
								echo "${volid}:$pvlun~$devblockpath" >> $idlistfile
							fi
						fi
					else
						if [ "$long" = "true" ]
						then
							echo "Skipping device $devblockpath because unsupported scsiid format: $scsiid"
						fi
						echo "Skipping device $devblockpath because unsupported scsiid format: $scsiid" >> $alogfile
					fi
				fi
			done
		done
		volumes=$(cat $idlistfile)
		echo "Volumes= $volumes" >> $alogfile
		rm -f $idlistfile
	fi

	let numtimes=0
	# Timeout loop - waiting for devices to appear
	while [ $numtimes -lt $innertries ]
	do
		let numtimes+=1
		echo "Begin volume wait loop number $numtimes..." >> $alogfile
		if [ "$long" = "true" ]
		then
			echo "Begin volume wait loop number $numtimes..."
		fi
		if [ "$firsttime" = "false" ]
		then
			echo "Pause 10..." >> $alogfile
			if [ "$long" = "true" ]
			then
				echo "Pause 10..."
			fi
			sleep 10
		fi
		firsttime="false"
		let numvols=0
		let numfound=0
		listnotfound=""
		if [ "$pvmapfile" != "" ]
		then
				rm -f $pvmapfile
		fi
		pvmapfile=$(mktemp $vmcdir/pvmap.XXXXXXXX)
		# Physical Volume loop - looking for physical volumes
		for vol in $volumes
		do
			echo "------------------------------" >> $alogfile
			echo "Volume: $vol" >> $alogfile
			if [ "$long" = "true" ]
			then
			echo "------------------------------"
			echo "Volume: $vol"
			fi
			let numvols+=1
			volfound=N
			mpath=N
			health=OK
			# Split volume arguments in pieces
			argwwid=${vol%%:*}
			argvoli=${vol##*:}
			arglun=${argvoli%%~*}
			argname=${argvoli##*~}
			if [ "$arglun" = "$argwwid" ]
			then
				arglun="-1"
				argname="[unknown-address]"
			fi
			if [ "$argname" = "$arglun" ]
			then
				argname="[unknown-address]"
			fi
			# Search in priority order
			if [ "$multipath" = "Y" ]
			then
				# 1. for multipath, search /dev/mapper
				stable=$(ls -l /dev/mapper | awk -v ss="[^p]$argwwid[^p]" '$0 ~ ss {print $9, $11}')
				if [ -n "$stable" ]
				then
					if [ "$long" = "true" ]
					then
						echo "stable: $stable"
					fi
					echo "stable: $stable" >> $alogfile
					dev=$(echo $stable | cut -d '/' -f 2)
					devpath=/dev/$dev
					stablename=$(echo $stable | cut -d ' ' -f 1)
					stablepath=/dev/mapper/$stablename
					volfound=Y
					mpath=Y
				else
					# fallback to by-id, though this should be pretty unlikely
					stable=$(ls -l /dev/disk/by-id | awk -v ss="dm-uuid-mpath-[0-9]$argwwid" '$0 ~ ss {print $9, $11}')
					if [ -n "$stable" ]
					then
						if [ "$long" = "true" ]
						then
							echo "stable: $stable"
						fi
						echo "stable: $stable" >> $alogfile
						dev=$(echo $stable | cut -d '/' -f 3)
						devpath=/dev/$dev
						stablename=$(echo $stable | cut -d ' ' -f 1)
						stablepath=/dev/mapper/$stablename
						volfound=Y
						mpath=Y
					fi
				fi
			fi
			if [ "$volfound" = "N" ]
			then
				stable=$(ls -l /dev/disk/by-id | awk -v ss="wwn-0x$argwwid[^-]" '$0 ~ ss {print $9, $11}')
				if [ -n "$stable" ]
				then
					if [ "$long" = "true" ]
					then
						echo "stable: $stable"
					fi
					echo "stable: $stable" >> $alogfile
					dev=$(echo $stable | cut -d '/' -f 3)
					devpath=/dev/$dev
					stablename=$(echo $stable | cut -d ' ' -f 1)
					stablepath=/dev/disk/by-id/$stablename
					volfound=Y

				fi
			fi
			if [ "$volfound" = "N" ]
			then
			    # Allow any scsi type prefix that is a single digit, not just '3' ('scsi-.' below)
				stable=$(ls -l /dev/disk/by-id | awk -v ss="scsi-.$argwwid[^-]" '$0 ~ ss {print $9, $11}')
				if [ -n "$stable" ]
				then
					if [ "$long" = "true" ]
					then
						echo "stable: $stable"
					fi
					echo "stable: $stable" >> $alogfile
					dev=$(echo $stable | cut -d '/' -f 3)
					devpath=/dev/$dev
					stablename=$(echo $stable | cut -d ' ' -f 1)
					stablepath=/dev/disk/by-id/$stablename
					volfound=Y
				fi
			fi
			if [ "$volfound" = "Y" ]
			then
				# Test device
				readpath=$(dd if=$devpath of=/dev/null bs=32 count=1 2>&1)
				rc=$?
				echo "dd if=$devpath of=/dev/null bs=32 count=1     [rc=$rc]" >> $alogfile
				if [ "$long" = "true" ]
				then
					echo "dd if=$devpath of=/dev/null bs=32 count=1     [rc=$rc]"
				fi
				if [ "$rc" != "0" ]
				then
					health="read-failure"
					# If the fix mode is "none", then no more volume 'fixing' will be attempted -
					# leave the volume as 'found' even though it cannot be read.
					if [ "$fixmode" != "none" ]
					then
						volfound=N
					fi
				fi
			fi
					
			if [ "$volfound" = "Y" ]
			then
				let numfound+=1
				# Common 'disk' pieces
				blockdir=/sys/block/$dev
				blocks=$(cat $blockdir/size)
				if [ "$blocks" = "" ]
				then
					blocks="0"
				fi
				blocksize=$(cat $blockdir/queue/physical_block_size)
				if [ "$blocksize" = "" ]
				then
					blocksize="0"
				fi
				# The last mult factor below is 1000 because sizes are specified in MB, so needed to match
				# Director allocations, but value is not used for anything...just eye-catcher
				sizeGiB=$(echo "scale=3; ($blocks * $blocksize) / ( 1024 * 1024 * 1000 )" | bc)
				ronly=$(cat $blockdir/ro)
				access=$ronly
				if [ "$ronly" = "0" ]
				then
					#read/write
					access="3"
				else
					if [ "$ronly" != "1" ]
					then
						#unknown
						access="0"
					fi
				fi

				if [ "$mpath" = "Y" ]
				then
					# For multipath output, assume the "active" path is 3 lines below the wwid entry.
					scsidev=$(multipath -ll | grep -A 3 $argwwid | tail -n 1 | awk '/running/ {print $3}')
					# The scsidev (e.g. 1:0:2:3) is used to index into the sysfs device path
					dlink=/sys/class/scsi_device/$scsidev/device
				else
					# Non-multipath case. The Device dir is right underneath the $blockdir
					dlink=$blockdir/device
					scsidev=$(ls $dlink/scsi_device)
				fi
				
				# Common 'device' pieces
				scsi3dev=${scsidev%:*}
				lun=$(echo $scsidev | cut -d : -f 4)
				vendor=$(cat $dlink/vendor)
				vendor=${vendor%%[ ]*}
				model=$(cat $dlink/model)
				model=${model%%[ ]*}
				transportdir=/sys/class/fc_transport/target$scsi3dev
				if [ -d $transportdir ]
				then
					hostwwpn=$(cat $transportdir/port_name)
				else
					# probably should exclude these from inventory
					hostwwpn="NON-FC-TRANSPORT"
				fi
				
				# send upper case wwid back
				deviceID=$(echo $argwwid | tr '[a-z]' '[A-Z]')
				# validation
				if [ "$hostwwpn" = "" ]
				then
					hostwwpn="unknown"
				fi
				if [ "$vendor" = "" ]
				then
					vendor="unknown"
				fi
				if [ "$model" = "" ]
				then
					model="unknown"
				fi
				
				# wwid  stable-path  devpath  scsi-addr  lun-in  host-lun  blocks  mpath  access  host-wwpn  vendor  model  external-address  block-size  sizeGiB  health
				if [ "$hostwwpn" != "NON-FC-TRANSPORT" ]
				then
					# Don't add this one to the return list if not an FC transport device
					printf '%-33s %70s %12s %11s %4s %4s %10s %1s %1s %18s %8s %8s %s %5s  %s  %s\n' $deviceID $stablepath $devpath $scsidev $arglun $lun $blocks $mpath "$access" $hostwwpn $vendor $model $argname $blocksize $sizeGiB $health >> $pvmapfile
				fi
				echo "  ------------- Found volume $numfound -------------------" >> $alogfile
				printf '%-33s %70s %12s %11s %4s %4s %10s %1s %1s %18s %8s %8s %s %5s  %s  %s\n' $deviceID $stablepath $devpath $scsidev $arglun $lun $blocks $mpath "$access" $hostwwpn $vendor $model $argname $blocksize $sizeGiB $health >> $alogfile
				if [ "$long" = "true" ]
				then
					echo "  ------------- Found volume $numfound -------------------"
					printf '%-33s %70s %12s %11s %4s %4s %10s %1s %1s %18s %8s %8s %s %5s  %s  %s\n' $deviceID $stablepath $devpath $scsidev $arglun $lun $blocks $mpath "$access" $hostwwpn $vendor $model $argname $blocksize $sizeGiB $health
				fi
			else
				if [ "$long" = "true" ]
				then
					echo "  **Volume not found: $argwwid  lun=$arglun  address=$argname"
				fi
				echo "  **Volume not found: $argwwid  lun=$arglun  address=$argname" >> $alogfile
				listnotfound="$listnotfound ${argwwid}:${arglun}~${argname}"
				#break (keep going instead?)
			fi
			# On to next volume to find
		done
		# End PV loop
		if [ "$numfound" = "$numvols" ]
		then
			fixmode="done"
			break
		fi
	done
	# End Timeout loop
	let innertries=3
	if [ "$fixmode" != "done" ]
	then
		if [ "$fixmode" = "id-enabled" ]
		then
			# run id fixup
			date >> $fixlogfile
			if [ "$argids" = "Y" ]
			then
				echo "  SAN check & fix: $vmcdir/sancheck1a.sh -f --ids $@" >> $alogfile
				echo "  SAN check & fix: $vmcdir/sancheck1a.sh -f --ids $@" >> $fixlogfile
				if [ "$long" = "true" ]
				then
					echo
					echo "  SAN check & fix: $vmcdir/sancheck1a.sh -f --ids $@"
				fi
				$vmcdir/sancheck1a.sh -f --ids $@ >> $fixlogfile
				rc=$?
				fixmode="full"
			else
				echo "  SAN check & fix: $vmcdir/sancheck1a.sh -fn --ids${listnotfound}" >> $alogfile
				echo "  SAN check & fix: $vmcdir/sancheck1a.sh -fn --ids${listnotfound}" >> $fixlogfile
				if [ "$long" = "true" ]
				then
					echo
					echo "  SAN check & fix: $vmcdir/sancheck1a.sh -fn --ids${listnotfound}"
				fi
				$vmcdir/sancheck1a.sh -fn --ids${listnotfound} >> $fixlogfile
				rc=$?
				fixmode="full"
				# skip full mode since we just did that equivalent as no ids were passed in
			fi
			echo "*** rc=$rc ***" >> $fixlogfile
			date >> $fixlogfile
		else
			if [ "$fixmode" = "full" ]
			then
				# run full fixup
				date >> $fixlogfile
				echo "  SAN check & fix: $vmcdir/sancheck1a.sh -f" >> $alogfile
				echo "  SAN check & fix: $vmcdir/sancheck1a.sh -f" >> $fixlogfile
				if [ "$long" = "true" ]
				then
					echo
					echo "  SAN check & fix: $vmcdir/sancheck1a.sh -f"
				fi
				$vmcdir/sancheck1a.sh -f >> $fixlogfile
				rc=$?
				echo "*** rc=$rc ***" >> $fixlogfile
				date >> $fixlogfile
				fixmode="none"
			else
				echo "Still could not find all the volumes. No more attmepts." >> $alogfile
				if [ "$long" = "true" ]
				then
					echo
					echo "Still could not find all the volumes. No more attmepts."
				fi
				fixmode="done"
			fi
		fi
	fi
done
# End outer fix loop

if [ "$numfound" != "$numvols" ]
then
	# Error case. We print extra things to stdout so that VMC logging will have some debug info.
	echo "All volume(s) NOT found!"
	echo "All volume(s) NOT found!" >> $alogfile
	if [ "$argids" = "N" ]
	then
		# return what we found anyway
		cat $pvmapfile
		cat $pvmapfile >> $alogfile
		sort -b -u --key=1,2 $pvmapfile $vmcdir/pvmap > $vmcdir/pvmap.1
		cp -f $vmcdir/pvmap.1 $vmcdir/pvmap
		rm -f $pvmapfile
		date >> $alogfile
		exit 0
	fi
	if [ -f $pvmapfile ]
	then
		echo "Dumping $pvmapfile ..."
		cat $pvmapfile
		rm -f $pvmapfile
	fi
	echo "-------------------------------------------------------"
	echo "Dumping sancheck table..."
	$vmcdir/sancheck1a.sh --ids $@
	echo "-------------------------------------------------------"
	echo "Dumping current pvmap list..."
	cat $vmcdir/pvmap
	echo "---------------------FAILED----------------------------"
	date
	date >> $alogfile
	exit 2
else
	# Success case
	echo "All volume(s) found!"
	echo "All volume(s) found!" >> $alogfile
	echo "wwid  stable-path  devpath  scsi-addr  lun-in  host-lun  blocks  mpath  access  host-wwpn  vendor  model  external-address  block-size  sizeGiB  health" >> $alogfile
	cat $pvmapfile
	cat $pvmapfile >> $alogfile
	sort -b -u --key=1,2 $pvmapfile $vmcdir/pvmap > $vmcdir/pvmap.1
	cp -f $vmcdir/pvmap.1 $vmcdir/pvmap
	rm -f $pvmapfile
	date >> $alogfile
	exit 0
fi
# unreachable
