#!/bin/bash
#* Licensed Materials - Property of IBM
#*
#* OCO Source Materials
#*
#* (C) Copyright IBM Corp. 2011 All Rights Reserved
#*
#*
# version: 1.1g   [1/05/12]
# author: csky
# sancheck1a.sh [-l | -f | -fn | -d | -h] [--ids <WWID>:<LUN>~<address> ...]
#    options: See the help text echoed out below.
#
# !Script only supported on RHEL 6!
#
# The devices that this script operates on are Fibre Channel SCSI devices.
# This version does not support virtual multipath devices.
#
# NOTE: the "STATUS" column on the short listing is a helpful tool, but does
#       not represet mutually exclusive volume states. Some of the status
#       categories chosen may overlap and one gets chosen. If the -f (fix)
#       option is specified, the STATUS is the state prior to any fix being done.
#       Here is a brief guide:
#   STATUS         DESCRIPTION
#   ------         -----------------------------------------------------------------
#   GOOD           Block-dev is readable and a WWID path is present in
#                  /etc/disk/by-id (stable path) for block-dev.
#   DUPED          Block-dev is readable, NO stable path, but there is a stable
#                  path to the same volume (different block-dev) that is readable.
#   PASSIVE        Block-dev is NOTE readable, but there is another block dev with
#                  a matching WWID that is readable.
#   PASSIVE-D      Block-dev is NOT readable, NO stable path, but there is a stable
#                  path to the same volume that is readable. (passived-duped)
#   PASSIVE-E      Block-dev is NOT readable, NO stable path, but there is a stable
#                  path to the same volume that is NOT readable. (passive-error)
#   PASSIVE-M      Block-dev is NOT readable, NO stable path, and no alternate
#                  stable paths to same volume. (passive-missing)
#   MISSING        Block-dev is readable, NO stable path, and no alternate stable
#                  paths to the same volume.
#   MISSING(rc)    Block-dev is readable, NO stable path, there is a stable path to
#                  the same volume that is NOT readable (rc=read return code).
#   BAD(-1)        The block-dev path does NOT even exist.
#   BAD(-2)        NO page 0x83 scsi-id could be retrieved for the block-dev.
#   BAD(-3)        Block-dev may or may not be readable, there is a stable path,
#                  but NO stable paths end in the WWID obtained.
#   BAD(rc)        Block-dev is not readable (rc=read error). Stale device.
#   OFFLINE        Block-dev is readable, but is empty (no data returned). Size
#                  is effectively '0'.
#   PREMATURE      Special bad state where there is a by-id 'stable path' but it
#                  begins with scsi-S<vendor> - which is not really stable, and not
#                  the WWID udev rule we require.
#
# Fixable states: BAD(x), PREMATURE, some 'partially bad' states, non-correlating symlinks.
#
long=false
if [ "$1" = "-h" -o "$1" = "--help" ]
then
	echo
	echo "Check FC SAN attached volumes."
	echo "Debug and scanning tool for IBM Systems Director VMControl."
	echo
	echo "sancheck1a.sh [-l | -f | -d | -h] [--ids <WWID>:<LUN> ...]"
	echo "  optional options:"
	echo "     -l      long dumping format. If neither -l or -f specified, dump is short table."
	echo "     -f      fix unusable block devices (forces long format)"
	echo "     -d      ensure removal of devices given by --ids (forces long format)"
	echo "     -h      help text"
	echo "     --ids   All arguments following this option are treated as a specific volume"
	echo "             reference, consisting of a WWID (DeviceID) and LUN combination"
	echo "             separated by a ':'.If specified, then only those devices are fully"
	echo "             processed. Missing stable paths can be fixed if the -f option is"
	echo "             specified in this case. If the LUN is unknown, pass -1."
	exit 0
fi
# Log file for posting run summaries
vmcdir="/var/opt/ibm/vmc"
logfile="$vmcdir/outsancheck.log"
fix=N
argids=N
del=N
if [ "$1" = "-l" ]
then
	long=true
fi
if [ "$1" = "-f" -o "$1" = "-fn" ]
then
	long=true
	fix=Y
fi
if [ "$1" = "-fn" ]
then
	limitrescan=Y
fi
if [ "$1" = "-d" ]
then
	del=Y
	long=true
fi
if [ "$1" = "--ids" ]
then
	shift 1
	argids=Y
else
	if [ "$2" = "--ids" ]
	then
		shift 2
		argids=Y
	fi
fi
if [ "$long" = "true" ]
then
	if [ -f $logfile ]
	then
		ccrc=$(find $logfile -size +4000)
		if [ "$ccrc" = "$logfile" ]
		then
			echo "Cleanup old log file (wrap)."
			rm -f $logfile
		fi
	fi
	date >> $logfile
	echo "fix=$fix  del=$del  argids=$argids" >> $logfile
	printf '%10s %7s %6s %10s %8s %34s %12s  %s\n' "SCSI-QUAD" "VENDOR" "MODEL" "BLK-DEV" "GB" "WWID" "STATE" "READ-CHECK" >> $logfile
	echo "**********************************************"
	echo "  KVM SAN SCSI volume consistency checking... "
	echo "**********************************************"
	echo
else
	echo "fix=$fix  del=$del  argids=$argids" 
	printf '%10s %7s %6s %10s %8s %34s %12s  %s\n' "SCSI-QUAD" "VENDOR" "MODEL" "BLK-DEV" "GB" "WWID" "STATE" "READ-CHECK"
fi
# Setup vars
let numdisks=0
let numbad=0
let numgood=0
let numremoved=0
let numadded=0
let nummissing=0
let numduped=0
let numpassive=0
let numoffline=0
let byidremoved=0
hosts=""

# Initially, make a table of all scsi devices to WWID mappings.
# The table is stored in a temp file
devmapfile=$(mktemp $vmcdir/devmap.XXXXXXXX)
for sysblk in $(ls -1 -d /sys/block/sd*)
do
	devname=${sysblk/sys\/block/dev}
	scsiid=$(scsi_id --page=0x83 --whitelisted --device=$devname)
	scsiid=${scsiid##[ ]*}
	if [ -n "$scsiid" ]
	then
		printf '%-13s %34s\n' "$devname" "$scsiid" >> $devmapfile
	fi
done
if [ "$long" = "true" ]
then
	mapped=$(cat $devmapfile | wc -l)
	echo "Created a wwid mapping file with $mapped devices: $devmapfile"
	cat $devmapfile
fi

# Check for FC trasport
if [ -d /sys/class/fc_transport ]
then
	if [ "$long" = "true" ]
	then
		echo "Detection of FC transport folder."
		echo
	fi
else
	if [ "$long" = "true" ]
	then
		echo "FC transport not detected - no volumes attached to check."
	fi
fi

# loop over transport combinations (HBAs, channels, controllers)
for fcdir in $(ls -1 /sys/class/fc_transport)
do
	if [ "$long" = "true" ]
	then
		echo "FC transport found: $fcfir"
	fi
	target=$(echo $fcdir | cut -d : -f 1)
	channel=$(echo $fcdir | cut -d : -f 2)
	controller=$(echo $fcdir | cut -d : -f 3)
	hostnum=${target#target}
	host=host${target#target}
	hosts=${hosts%%$host}
	if [ -z "$hosts" ]
	then
		hosts=$host
	else
		hosts="$hosts,$host"
	fi
	portid=$(cat /sys/class/fc_transport/$fcdir/port_id)
	portwwn=$(cat /sys/class/fc_transport/$fcdir/node_name)
	hportid=$(cat /sys/class/fc_host/$host/port_id)
	hportwwn=$(cat /sys/class/fc_host/$host/port_name)
	if [ "$long" = "true" ]
	then
		echo "  target=$target  channel=$channel  controller=$controller  host=$host  hostnum=$hostnum"
		echo "  SAN_WWNN:   $portwwn    PortID:  $portid"
		echo "  Host_WWNN:  $hportwwn    PortID:  $hportid"
	fi
	# TODO libvirt virsh stuff here if corresponding managed volumes found
	
    for file in $(ls -1 /sys/class/fc_transport/$fcdir/device | egrep "^$hostnum:$channel:$controller:")
    do
		dlink=/sys/class/fc_transport/$fcdir/device/$file
		type=$(cat $dlink/type)
		state=$(cat $dlink/state)
		lun=$(echo $file | cut -d : -f 4)
		vendor=$(cat $dlink/vendor)
		vendor=${vendor%%[ ]*}
		model=$(cat $dlink/model)
		model=${model%%[ ]*}
		blockdir=""
		devcheck="unknown"
		blockdev=""
		size=""
		sizegb=""
		ronly=""
		devmajor=""
		devminor=""
		capability=""
		devblockpath=""
		devpathstate=""
		dstate=""
		if [ -d $dlink/block ]
		then
			blockdev=$(ls $dlink/block)
			blockdir=$dlink/block/$blockdev
			size=$(cat $blockdir/size)
			sizegb=$(echo "scale=3; $size * 512 / ( 1024 * 1024 * 1000 )" | bc)
			ronly=$(cat $blockdir/ro)
			devmajor=$(cat $blockdir/dev | cut -d : -f 1)
			devminor=$(cat $blockdir/dev | cut -d : -f 2)
			capability=$(cat $blockdir/capability)
			devblockpath=/dev/$blockdev
			devpathstate="Missing"
			if [ -b $devblockpath ]
			then
				devpathstate="Exists"
			fi
			if [ "$ronly" = "0" ]
			then
				ronly=false
			else
				ronly=true
			fi
		else
			blockdev="n/a"
		fi
		if [ "$long" = "true" ]
		then
			echo
			echo "======================================================================="
			echo "    Examining FC device: $dlink"
			echo "      LUN               $lun"
			echo "      BlockDevice:      $blockdev"
			echo "      Type:             $type"
			echo "      DevState:         $state"
			echo "      Size:             $sizegb GB"
			echo "      ReadOnly:         $ronly"
			echo "      DeviceMajor:      $devmajor"
			echo "      DeviceMinor:      $devminor"
			echo "      Capability:       $capability"
			echo "      Vendor:           $vendor"
			echo "      Model:            $model"
			echo "      BlockFilePath:    $devblockpath"
			echo "      BlockSpecialFile: $devpathstate"
			echo
		fi
		if [ -z $blockdir ]
		then
			if [ "$long" = "true" ]
			then
				echo "    ***Non block device, so skip to next FC device..."
			fi
			continue
		fi
		let numdisks=$numdisks+1
		scsiid='[unknown]'
		volid='[unknown]'
		idmatch='false'
		checkrc="0"
		inputmatch=N
		if [ "$devpathstate" = "Exists" ]
		then
			scsiid=$(scsi_id --page=0x83 --whitelisted --device=$devblockpath)
			scsiid=${scsiid##[ ]*}
			# remove first char (3) from scsi ID to get the backend VOL ID.
			volid=${scsiid:1}
			volid=${volid##[ ]*}
			if [ "$scsiid" = "" -o "$volid" = "" ]
			then
				volid='[cannot retrieve]'
				# No WWID case
				checkrc="-2"
			else
				# Check against any input ids
				if [ "$argids" = "Y" ]
				then
					for argid in $@
					do
						argwwid=${argid%%:*}
						argvoli=${argid##*:}
						arglun=${argvoli%%~*}
						argname=${argvoli##*~}
						if [ "$arglun" = "$argwwid" ]
						then
							arglun="-1"
							argname="[unknown-address]"
						fi
						if [ "$argname" = "$arglun" ]
						then
							argname="[unknown-address]"
						fi
						# The wwid is expected in the proper alpha case
						if [ "$argwwid" = "$volid" ]
						then
							inputmatch=Y
							if [ "$long" = "true" ]
							then
								echo "    Found input volume match: $argid"
								if [ "$arglun" != "-1" -a "$arglun" != "$lun" ]
								then
									echo "      *Input LUN (external) is not a match!!"
								fi
							fi
							if [ "$del" = "Y" ]
							then
								# remove this device - special tag
								checkrc="DEL"
								echo "!!Mark device for deletion..."
							fi
						fi
					done
				fi
			fi
			if [ "$long" = "true" ]
			then
				echo "    Vol SCSI ID:   $scsiid"
				echo "    Trimmed VolID:  $volid"
				echo "    Do block device I/O check on existing block special file..."
			fi
			# The check is done even if the WWID is not there
			if [ "$argids" = "N" -o "$inputmatch" = "Y" ]
			then
				devcheck=$(file -b -s $devblockpath)
				rc=$?
			else
				devcheck=$(dd if=$devblockpath of=/dev/null bs=32 count=1 2>&1)
				rc=$?
				if [ "$rc" = "0" ]
				then
					devcheck=$(echo "$devcheck" | head -n 1)
				fi
			fi
			if [ "$long" = "true" ]
			then
				echo "      file/dd read check $devblockpath: rc=$rc"
				echo "        Output: $devcheck"
			fi
			if [ "$devcheck" = "empty" ]
			then
				let numoffline=$numoffline+1
				if [ "$long" = "true" ]
				then
					echo "        ***No data to read. Marking storage as offline."
				fi
			fi
			if [ "$checkrc" = "0" ]
			then
				checkrc=$rc
			fi
		else
			checkrc="-1"
		fi
		fixthis=N
		if [ "$fix" = "Y" -a "$argids" = "N" ]
		then
			fixthis=Y
		else
			if [ "$fix" = "Y" -a "$inputmatch" = "Y" ]
			then
				fixthis=Y
			fi
		fi
		# If any 'check' failure at this point, remove the line from the mapping file
		# So that it is not mistaken for a good device later and checked.
		if [ "$checkrc" != 0 ]
		then
			preln=$(cat $devmapfile | wc -l)
			sed -i "/${blockdev}[ ]/d" $devmapfile
			postln=$(cat $devmapfile | wc -l)
			if [ "$long" = "true" ]
			then
					echo "    sed -i /${blockdev}[ ]/d $devmapfile"
					echo "    Removed $blockdev row from $devmapfile. $preln lines to $postln."
			fi
		fi

		# correlate to by-id
		let numbyids=0
		if [ "$blockdev" != "n/a" ]
		then
			if [ "$long" = "true" ]
			then
				echo
				echo "    /dev/disk/byi-id correlation(s):"
			fi
			badmatch="N"
			for byident in $(ls -l /dev/disk/by-id | egrep "/$blockdev\$" | awk '{print $9}')
			do
				# Does byident end with volid?
				idchk=${byident%$volid}
				if [ "$idchk" != "$byident" ]
				then
				    # Since not equal, the volid portion was removed and symlink matches
					idmatch='true'
					let numbyids=$numbyids+1
					prematcheck=${byident#scsi-S$vendor}
					if [ "$prematcheck" != "$byident" ]
					then
						let numpremat=$numpremat+1
						dstate="PREMATURE"
						if [ "$long" = "true" ]
						then
							echo "    PREMATURE stable path link name: $byident"
						fi
					fi
				else
					if [ "$fixthis" = "Y" ]
					then
						# cleanup by-id link & rescan
						rm -f /dev/disk/by-id/$byident
						rmret=$?
						echo "  **Cleanup bad link: /dev/disk/by-id/$byident    RC=$rmret"
					fi
					# Only if device is unreadable, mark as a bad match.
					if [ "$checkrc" = "1" ]
					then
						badmatch="Y"
					fi
				fi
				if [ "$long" = "true" ]
				then
					echo "          $byident     (ID-Match = $idmatch)"
				fi
			done
			if [ "$numbyids" = "0" -a "$badmatch" = "Y" ]
			then
				# If the scsi block dev already exists, rescanning won't help
				# e.g. it won't cause a new /dev/sdXX and/or a new link in 'by-id', so mark for cleanup
				if [ "$long" = "true" ]
				then
					echo "    Mark device for cleanup: $blockdev"
				fi
				checkrc="-3"
			fi
		fi
		if [ "$long" = "true" ]
		then
			echo "    CheckRC=$checkrc"
		fi
		if [ "$checkrc" = "1" ]
		then
			# check for passive link case. Need to loop through all scsi devices looking for matching ids
			for chkdev in $(grep $scsiid $devmapfile | awk '{print $1}')
			do
				if [ "$chkdev" != "$devblockpath" ]
				then
					devcheck2=$(dd if=$chkdev of=/dev/null bs=32 count=1 2>&1)
					rc=$?
					if [ "$rc" = "0" ]
					then
						if [ "$dstate" = "" ]
						then
							dstate="PASSIVE"
							checkrc="0"
						fi
						let numpassive=$numpassive+1
						if [ "$long" = "true" ]
						then
							echo "    Another active link found to volume: $chkdev    So $blockdev is passive. Cur state: $dstate"
						fi
						break
					fi
				fi
			done
		fi
		if [ "$checkrc" = "0" -a "$dstate" = "PREMATURE" ]
		then
			# a premature state where no other issue overrides it
			checkrc="-4"
		fi

		# Final judgement
		if [ "$checkrc" = "0" ]
		then
			if [ "$numbyids" != "0" ]
			then
				if [ "$dstate" != "PASSIVE" ]
				then
					if [ "$devcheck" = "empty" ]
					then
						dstate="OFFLINE"
						# FIX: This is not a fixable condition since it indicates a backend storage issue
					else
						let numgood=$numgood+1
						dstate="GOOD"
						if [ "$long" = "true" ]
						then
							echo "    ---GOOD---"
						fi
					fi
				else
					if [ "$long" = "true" ]
					then
						echo "    ---PASSIVE-GOOD---"
					fi
				fi
			else
				# Another block device to the same volume might have a link in by-id though.
				byidww=$(ls -l /dev/disk/by-id | awk -v ss="wwn-0x$volid[^-]" '$0 ~ ss {print $9, $11}')
				if [ -z "$byidww" ]
				then
					 # Allow any scsi type prefix that is a single digit, not just '3' ('scsi-.' below)
					byidww=$(ls -l /dev/disk/by-id | awk -v ss="scsi-.$volid[^-]" '$0 ~ ss {print $9, $11}')
				fi
				if [ -z "$byidww" ]
				then
					if [ "$long" = "true" ]
					then
						echo "    ***WWID BY-ID PATH MISSING***"
					fi
					if [ "$dstate" = "PASSIVE" ]
					then
						dstate="PASSIVE-M"
					else
						# Could be corrected if there was a link in 'by-id'
						dstate="MISSING"
						let nummissing=$nummissing+1
						if [ "$fix" = "Y" -a "$inputmatch" = "Y" ]
						then
							# make a symlink. None currently exist.
							linkname="/dev/disk/by-id/wwn-0x$volid"
							ln -s $devblockpath $linkname
							rc=$?
							if [ "$rc" = "0" ]
							then
								let numadded=$numadded+1
							fi
							echo "    Adding stable symlink: ln -s $devblockpath $linkname   RC=$rc"
						fi
					fi
				else
					otherdev=$(echo $byidww | cut -d '/' -f 3)
					dev2check=$(file -b -s /dev/$otherdev)
					rc=$?
					if [ "$long" = "true" ]
					then
						echo "    Nothing in /dev/disk/by-id points to block dev $blockdev, but there is a link for same vol:"
						echo "      $byidww"
						echo "    Check: file -b -s /dev/$otherdev --> rc=$rc  output=$dev2check"
						if [ "$rc" = "0" ]
						then
							echo "    $dstate ---DUPLICATE---"
						else
							echo "    ***WWID BY-ID PATH MISSING (bad symlink) $dstate***"
							# Nothing to cleanup here since this is a different device - needs to be caught in another loop
						fi
					fi
					if [ "$rc" = "0" ]
					then
						if [ "$dstate" = "PASSIVE" ]
						then
							dstate="PASSIVE-D"
						else
							dstate="DUPED"
							let numduped=$numduped+1
						fi
					else
						if [ "$dstate" = "PASSIVE" ]
						then
							dstate="PASSIVE-E"
						else
							dstate="MISSING($rc)"
							let nummissing=$nummissing+1
							# fix possible here too but old link(s) must be removed
							if [ "$fix" = "Y" -a "$inputmatch" = "Y" ]
							then
								linkname="/dev/disk/by-id/wwn-0x$volid"
								if [ -L "$linkname" ]
								then
									rm -f $linkname
									rc=$?
									echo "      REMOVE: $linkname  (rc=$rc)"
									if [ "$rc" = "0" ]
									then
										let byidremoved=$byidremoved+1
									fi
								fi
								ln -s $devblockpath $linkname
								rc=$?
								if [ "$rc" = "0" ]
								then
									let numadded=$numadded+1
								fi
								echo "    Adding stable symlink: ln -s $devblockpath $linkname   RC=$rc"
							fi
						fi
					fi
				fi
			fi
		else
			let numbad=$numbad+1
			# check for special premature state
			if [ "$dstate" != "PREMATURE" ]
			then
				dstate="BAD($checkrc)"
			fi
			if [ "$checkrc" = "DEL" ]
			then
				dstate="DELETED*"
				echo 1 > /sys/class/scsi_device/$file/device/delete
				rc=$?
				echo "!!Remove device: echo 1 > /sys/class/scsi_device/$file/device/delete  (rc=$rc)"
				if [ "$rc" = "0" ]
				then
					let numremoved=$numremoved+1
				fi
			fi
			if [ "$long" = "true" ]
			then
				echo "    ***BAD/STALE*** rc=$checkrc"
			fi
			if [ "$fixthis" = "Y" ]
			then
				# Fix: remove the 'bad' device
				echo "    Removing stale device $devblockpath (echo 1 > $dlink/delete) ..."
				echo 1 > $dlink/delete
				rc=$?
				echo "    rc=$?"
				if [ "$rc" = "0" ]
				then
					let numremoved=$numremoved+1
				fi
			else
				if [ "$long" = "true" ]
				then
					echo "    Fix stale links not requested. Continue..."
				fi
			fi
		fi
		# table dumper
		content=${devcheck:0:42}
		smodel=${model:0:5}
		if [ "$inputmatch" = "Y" ]
		then
			dstate="${dstate}*"
		fi

		if [ "$long" != "true" ]
		then
			# SCSI-QUAD  VENDOR  MODEL  BLOCK  GB  WWID  STATE  READ-CHECK
			printf '%10s %7s %6s %10s %8.3f %34s %12s  %s\n' "$file" "$vendor" "$smodel" "$devblockpath" "$sizegb" "$volid" "$dstate" "$content"
		else
			printf '%10s %7s %6s %10s %8.3f %34s %12s  %s\n' "$file" "$vendor" "$smodel" "$devblockpath" "$sizegb" "$volid" "$dstate" "$content" >> $logfile
		fi
    done
	if [ "$long" = "true" ]
	then
		echo "==========================================================================="
		echo "==========================================================================="
	fi
done

if [ -z "$hosts" ]
then
	echo "There does not appear to be any SAN volumes attached to the host."
	exit -1
else
	# Also manually clean up stale links in /dev/disk/by-id
	if [ "$fix" = "Y" -a "$argids" = "N" -o "$del" = "Y" ]
	then
		echo
		if [ "$numremoved" != "0" ]
		then
			echo "idling 5 seconds..."
			sleep 5
		fi
		echo "Symlink checks in by-id..."
		for symlink in $(ls -1 /dev/disk/by-id | grep "wwn-0x")
		do
			device=$(ls -l /dev/disk/by-id/$symlink | awk '{print $11}' | cut -d '/' -f 3)
			checkout=$(dd if=/dev/$device of=/dev/null bs=32 count=1 2>&1)
			rc=$?
			if [ "$rc" = "0" ]
			then
				echo "  Checked OK: /dev/disk/by-id/$symlink -> /dev/$device"
			else
				rm -f /dev/disk/by-id/$symlink
				rc2=$?
				echo "  REMOVED   : /dev/disk/by-id/$symlink -> /dev/$device  (read-rc=$rc, rm-rc=$rc2)"
				if [ "$rc2" = "0" ]
				then
					let byidremoved=$byidremoved+1
				fi
			fi
		done
		for symlink in $(ls -1 /dev/disk/by-id | grep "scsi-")
		do
			device=$(ls -l /dev/disk/by-id/$symlink | awk '{print $11}' | cut -d '/' -f 3)
			checkout=$(dd if=/dev/$device of=/dev/null bs=32 count=1 2>&1)
			rc=$?
			if [ "$rc" = "0" ]
			then
				echo "  Checked OK: /dev/disk/by-id/$symlink -> /dev/$device"
			else
				rm -f /dev/disk/by-id/$symlink
				rc2=$?
				echo "  REMOVED   : /dev/disk/by-id/$symlink -> /dev/$device  (read-rc=$rc, rm-rc=$rc2)"
				if [ "$rc2" = "0" ]
				then
					let byidremoved=$byidremoved+1
				fi
			fi
		done
		#for symlink in $(ls -1 /dev/disk/by-id | grep "scsi-SIBM_")
		#do
		#	rm -f /dev/disk/by-id/$symlink
		#	rc=$?
		#	echo "  REMOVED premature link: /dev/disk/by-id/$symlink   ($rc)"
		#	let byidremoved=$byidremoved+1
		#done
		
		let totalremoved=$numremoved+$byidremoved
		if [ "$totalremoved" != "0" -a "$del" = "N" -a "$limitrescan" != "Y" ]
		then
			# Removal took place -- rescan?
			echo
			echo "Since cleanup took place, rescan...."
			sleep 2
			for rescan in $(echo $hosts | sed s/,/\\n/g)
			do
				echo "- - -" > /sys/class/scsi_host/$rescan/scan
				echo "   echo '- - -'  > /sys/class/scsi_host/$rescan/scan"
			done
		fi
	fi
	if [ "$long" = "true" ]
	then
		echo
		echo "Host Bus Adapter list: $hosts"
		echo "Number volumes: $numdisks" >> $logfile
		echo "Number volumes: $numdisks"
		echo "     Bad/stale: $numbad    Good: $numgood    Missing: $nummissing    Passsive: $numpassive    Dups: $numduped    Offline: $numoffline" >> $logfile
		echo "     Bad/stale: $numbad    Good: $numgood    Missing: $nummissing    Passsive: $numpassive    Dups: $numduped    Offline: $numoffline"
		echo "     Stale devices removed: $numremoved      Links removed: $byidremoved      Stable links added: $numadded" >> $logfile
		echo "     Stale devices removed: $numremoved      Links removed: $byidremoved      Stable links added: $numadded"
	else
		echo
		echo "Number volumes: $numdisks"
		echo "     Bad/stale: $numbad    Good: $numgood    Missing: $nummissing    Passsive: $numpassive    Dups: $numduped    Offline: $numoffline"
	fi
	if [ "$long" = "true" ]
	then
		date >> $logfile
		echo "Done."
	fi
fi
rm -f $devmapfile
let totalfixed=$totalremoved+$numadded
exit $totalfixed
